#include "bm_load.h"

//#define LOADELF_DEBUG

unsigned char elfIdent[0x10] = { 0x7f, 0x45, 0x4c, 0x46, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

int loadElf(char *filename, int *entryvar)
{
	struct elfHeader elfHeader;
	struct elfProgHeader *progHeader;
	int progHeaderSize;
	int fd, fdSize;
	int i;

	fd = fioOpen(filename, O_RDONLY);
	if(fd < 0) return ERROR_LOADELF_FILEIO;

	fdSize = fioLseek(fd, 0, SEEK_SET);
	fioLseek(fd, 0, SEEK_SET);

	if(fioRead(fd, &elfHeader, sizeof(struct elfHeader)) != sizeof(struct elfHeader))
		return ERROR_LOADELF_FILEIO;

	// Check for valid elf
	if(memcmp(elfHeader.e_ident, elfIdent, 16) != 0)
		return ERROR_LOADELF_ELF;

	*entryvar = elfHeader.e_entry;
#ifdef LOADELF_DEBUG
	printf("Entry point: 0x%X\n", *entryvar);
#endif

	// Sanity checks
	if(!elfHeader.e_phoff) return ERROR_LOADELF_ELF;
	if(elfHeader.e_phentsize != sizeof(struct elfProgHeader)) return ERROR_LOADELF_ELF;
	
	// Allocate memory for program header table, read in
	progHeaderSize = elfHeader.e_phentsize * elfHeader.e_phnum;
	progHeader = (struct elfProgHeader *)malloc(progHeaderSize);
	if(!progHeader) return ERROR_LOADELF_MEMORY;

	fioLseek(fd, elfHeader.e_phoff, SEEK_SET);
	if(fioRead(fd, progHeader, progHeaderSize) != progHeaderSize)
		return ERROR_LOADELF_FILEIO;

	// Iterate over each program header entry, load into ram if correct type
	for(i = 0; i < elfHeader.e_phnum; i++)
	{
		if(progHeader[i].p_type == PT_LOAD)
		{
#ifdef LOADELF_DEBUG
		printf("Loading segment %d: offset = 0x%x, virtaddr = 0x%X, filesize = 0x%X, memsize = 0x%X\n",
				i, progHeader[i].p_offset, progHeader[i].p_vaddr, progHeader[i].p_filesz, progHeader[i].p_memsz);
#endif

		// Make sure segment to be loaded wont overwrite our elf
		if(	(progHeader[i].p_vaddr < (int)&_stack_end) && 
			((progHeader[i].p_vaddr + progHeader[i].p_memsz) > (int)&__start) )
			return ERROR_LOADELF_OVERWRITE;

		// Seet to offset within elf file
		fioLseek(fd, progHeader[i].p_offset, SEEK_SET);

		// Read segment from elf file into memory
		if(fioRead(fd, (void *)progHeader[i].p_vaddr, progHeader[i].p_filesz) != progHeader[i].p_filesz)
			return ERROR_LOADELF_FILEIO;

		// Zero extra space if needed
		if(progHeader[i].p_filesz != progHeader[i].p_memsz)
			memset((void *)(progHeader[i].p_vaddr + progHeader[i].p_filesz), 0, 
							progHeader[i].p_memsz - progHeader[i].p_filesz);
		}
	}

	fioClose(fd);
	free(progHeader);

	return 0;
}
