#ifndef _CDVDFSV_H
#define _CDVDFSV_H

// Macros for READ Data pattan
#define CdSecS2048		0	// sector size 2048
#define CdSecS2328		1	// sector size 2328
#define CdSecS2340		2	// sector size 2340

// Macros for Spindle control
#define CdSpinMax		0
#define CdSpinNom		1	
#define CdSpinStm		0	

// Macros for Error code
#define CdErFAIL     -1
#define CdErNO		0x00
#define CdErEOM		0x32
#define CdErTRMOPN	0x31
#define CdErREAD		0x30
#define CdErPRM		0x22
#define CdErILI		0x21
#define CdErIPI		0x20
#define CdErCUD		0x14
#define CdErNORDY	0x13
#define CdErNODISC	0x12
#define CdErOPENS	0x11
#define CdErCMD		0x10
#define CdErABRT		0x01
#define CdErREADCF	0xfd

// Macros for CdGetDiskType()
#define CdIllgalMedia 	0xff	// Illegal media
#define CdDVDV			0xfe	// DVD video
#define CdCDDA			0xfd	// CD-DA
#define CdPS2DVD		0x14	// PS2 DVD disc
#define CdPS2CDDA		0x13	// PS2 CD-DA disc
#define CdPS2CD			0x12	// PS2 CD disc
#define CdPSCDDA 		0x11	// PSX CD-DA disc
#define CdPSCD			0x10	// PSX CD disc
#define CdDETCT			0x01	// Detecting
#define CdNODISC 		0x00	// No disc
#define CdUNKNOWN		0x05	// Unknown disc

// CD-ROM Primitive Commands


// Interrupts
#define CdlNoIntr		0x00	// No interrupt		  
#define CdlDataReady	0x01	// Data Ready		  
#define CdComplete	0x02	// Command Complete 	  
#define CdlAcknowledge	0x03	// Acknowledge (reserved) 
#define CdlDataEnd		0x04	// End of Data Detected   
#define CdlDiskError	0x05	// Error Detected 	  
#define CdNotReady	0x06	// Drive Not Ready	  

// Status Contents
#define CdStatShellOpen      0x01    // once shell open 	 
#define CdStatStop	        0x00    // Stop		 	  
#define CdStatSpin           0x02    // spindle motor rotating 
#define CdStatRead           0x06    // reading data sectors   
#define CdStatPause          0x0a    // pause		  
#define CdStatSeek           0x12    // seeking 		  
#define CdStatEmg	        0x20    // error detected	  

// Library Macros
#ifndef btoi
#define btoi(b)		((b)/16*10 + (b)%16)		// BCD to u8
#endif
#ifndef itob
#define itob(i)		((i)/10*16 + (i)%10)		// u8 to BCD
#endif

// Position
#define CdlMAXTOC	100

// Callback Reason
#define CdFuncRead		1
#define CdFuncSeek		4
#define CdFuncStandby	5
#define CdFuncStop		6
#define CdFuncPause		7
#define CdFuncBreak      8


typedef void (*CdlCB)(u8,u8 *);


// CD and DVD Read Mode
typedef struct {
	u8 trycount;	// trycount   
	u8 spindlctrl;	// spindlctrl 
	u8 datapattern;	// datapattern 
	u8 pad;			// pad        
} CdRMode;


// Location
typedef struct {
	u8 minute;		// minute (BCD)
	u8 second;		// second (BCD) 
	u8 sector;		// sector (BCD) 
	u8 track;		// track (void) 
} CdlLOCCD;

// CdInit Mode
#define CdInitChk		0x00    // disk chk	   
#define CdInitNoChk		0x01    // disk no chk    
#define CdExit			0x05    // cdvd exit	   

// Low Level File System for CdSearchFile() 
#define CdlMAXFILE	64		// max number of files in a directory 
#define CdlMAXDIR	128		// max number of total directories 
#define CdlMAXLEVEL	8		// max levels of directories 

typedef struct {
	u32 lsn;			// file location 
	u32	size;			// file size 
	char name[16];		// file name (body) 
	u8 date[8];			// date		    
    u32 flag;			// file_flag        
} CdlFILE;


// RTC
typedef struct {
        u8 stat;            // status 
        u8 second;          // second 
        u8 minute;          // minute 
        u8 hour;            // hour   

        u8 pad;             // pad    
        u8 day;             // day    
        u8 month;           // month  
        u8 year;            // year   
} CdCLOCK;

// Stream

//  Stream mode
#define STMNBLK         0
#define STMBLK          1

// Stream Init Devctl()
typedef struct {
        u32 bufmax;
        u32 bankmax;
        u32 iop_bufaddr;
} CdStmInit;

// Media mode
#define CdCD         1
#define CdDVD        2

// Macros for CdTrayReq
#define CdTrayOpen	0	// Tray Open  
#define CdTrayClose	1	// Tray Close 
#define CdTrayCheck	2	// Tray Check 

// EE Read mode
#define CdNoCheckReady       0x00000001
#define CdNoWriteBackDCache  0x00000002

// Prototypes

// DONE AND TESTED
int CdInit(int init_mode);
int CdStop(void);
int CdTrayReq(int param,u32 *traychk);
int CdDiskReady(int mode);
int CdGetDiskType(void);
int CdGetToc(u8 *toc);
int CdSearchFile(CdlFILE *fp, const char *name);
int CdRead(u32 lbn, u32 sectors, void *buf , CdRMode *mode);
int CdStandby(void);
int CdReadKey(int unk1, int unk2, int unk3, unsigned char *key);

// DONE

// TODO

CdlLOCCD *CdIntToPos(int i, CdlLOCCD *p);
int CdPosToInt(CdlLOCCD *p);

int CdReadIOPm(u32 lbn, u32 sectors, void *buf , CdRMode *mode);
int *CdCallback(void (*func)(int));
u32 CdGetReadPos( void );
int CdSync(int mode);
int CdInitEeCB(int cb_prio, void *stack_addr,int stack_size);
int CdGetError(void);
int CdStatus(void);
int CdBreak(void);
int CdSeek(u32 lbn);
int CdPause(void);
int CdStInit(u32 bufmax, u32 bankmax,  u32 iop_bufaddr);
int CdStStart(u32 lbn, CdRMode *mode);
int CdStSeek(u32 lbn);
int CdStStop(void);
int CdStRead(u32 size, u32 *buf, u32 mode, u32 *err);
int CdStPause(void);
int CdStResume(void);
int CdStStat(void);
int CdReadClock(CdCLOCK *rtc);

#endif
