#include <tamtypes.h>
#include <string.h>
#include <kernel.h>
#include <stdio.h>
#include <stdlib.h>
#include <sifrpc.h>
#include <stdarg.h>

#include "cdvdfsv.h"

#define CD_SERVER_INIT			0x80000592
#define CD_SERVER_SCMD			0x80000593
#define CD_SERVER_NCMD			0x80000595
#define CD_SERVER_DISKREADY		0x8000059A
#define CD_SERVER_SEARCHFILE	0x80000597

#define CD_SCMD_DISKTYPE	0x03
#define CD_SCMD_TRAYREQ		0x05

#define CD_NCMD_READ		0x01
#define CD_NCMD_GETTOC		0x04
#define CD_NCMD_STANDBY		0x06
#define CD_NCMD_STOP		0x07
#define CD_NCMD_READKEY		0x0B

static unsigned int rpcBuffer[512] __attribute__((aligned (64))); // May need to make this bigger
static unsigned char intrData[192] __attribute__((aligned (64))); // Dunno about these array sizes, but they will do
static unsigned int readCurPos[4] __attribute__((aligned (64)));

static struct t_RpcClientData CdClientInit __attribute__((aligned (64)));
static struct t_RpcClientData CdClientScmd __attribute__((aligned (64))); 
static struct t_RpcClientData CdClientNcmd __attribute__((aligned (64))); 
static struct t_RpcClientData CdClientDiskReady __attribute__((aligned (64)));
static struct t_RpcClientData CdClientSearchFile __attribute__((aligned (64)));

static int CdInited = 0;

// Stuff for CdSearchFile
struct SearchFilePkt {
	char	padding[32];
	char	name[256];
	void	*dest;
};
static CdlFILE SearchFileRecv __attribute__((aligned (64)));

// Stuff for CdRead
struct ReadPkt {
	u32 lbn;
	u32 sectors;
	void *buffer;
	CdRMode ReadMode;
	void *intr_data;
	void *read_pos;
};

int CdInit(int init_mode)
{
	int i;

	if(CdInited) return 0;

	while(1)
	{
		if (SifBindRpc(&CdClientInit, CD_SERVER_INIT, 0) < 0) return -1;
 		if (CdClientInit.server != 0) break;

    	i = 0x10000;
    	while(i--);
	}

	while(1)
	{
		if (SifBindRpc(&CdClientScmd, CD_SERVER_SCMD, 0) < 0) return -1;
 		if (CdClientScmd.server != 0) break;

    	i = 0x10000;
    	while(i--);
	}

	while(1)
	{
		if (SifBindRpc(&CdClientNcmd, CD_SERVER_NCMD, 0) < 0) return -1;
 		if (CdClientNcmd.server != 0) break;

    	i = 0x10000;
    	while(i--);
	}

	while(1)
	{
		if (SifBindRpc(&CdClientDiskReady, CD_SERVER_DISKREADY, 0) < 0) return -1;
 		if (CdClientDiskReady.server != 0) break;

    	i = 0x10000;
    	while(i--);
	}

	while(1)
	{
		if (SifBindRpc(&CdClientSearchFile, CD_SERVER_SEARCHFILE, 0) < 0) return -1;
 		if (CdClientSearchFile.server != 0) break;

    	i = 0x10000;
    	while(i--);
	}

	rpcBuffer[0] = init_mode;
	SifCallRpc(&CdClientInit, 0, 0, rpcBuffer, 4, rpcBuffer, 4, 0, 0);

	if(rpcBuffer[0]) CdInited = 1;

	return rpcBuffer[0];
}

int CdStop(void)
{
	if(!CdInited) return -1;

	SifCallRpc(&CdClientNcmd, CD_NCMD_STOP, 0, rpcBuffer, 0, rpcBuffer, 4, 0, 0);

	return rpcBuffer[0];
}

int CdTrayReq(int param, u32 *traychk)
{
	if(!CdInited) return -1;

	rpcBuffer[0] = param;
	SifCallRpc(&CdClientScmd, CD_SCMD_TRAYREQ, 0, rpcBuffer, 4, rpcBuffer, 8, 0, 0);

	if(rpcBuffer[0]) *traychk = rpcBuffer[1];

	return rpcBuffer[0];
}

int CdDiskReady(int mode)
{
	if(!CdInited) return -1;

	rpcBuffer[0] = mode;
	SifCallRpc(&CdClientDiskReady, 0, 0, rpcBuffer, 4, rpcBuffer, 4, 0, 0);

	return rpcBuffer[0];
}

int CdGetDiskType(void)
{
	if(!CdInited) return -1;

	SifCallRpc(&CdClientScmd, CD_SCMD_DISKTYPE, 0, rpcBuffer, 0, rpcBuffer, 4, 0, 0);

	return rpcBuffer[0];
}

int CdGetToc(u8 *toc)
{
	if(!CdInited) return -1;

	rpcBuffer[0] = (unsigned int)toc;

	if (!IS_UNCACHED_SEG(toc))
		SifWritebackDcache(toc, 2064);

	SifCallRpc(&CdClientNcmd, CD_NCMD_GETTOC, 0, rpcBuffer, 4, rpcBuffer, 4, 0, 0);

	return rpcBuffer[0];
}

int CdSearchFile(CdlFILE *fp, const char *name)
{
	struct SearchFilePkt *searchFile = (struct SearchFilePkt *)rpcBuffer;

	if(!CdInited) return -1;

	strncpy(searchFile->name, name, 256);
	searchFile->name[255] = '\0';

	searchFile->dest = &SearchFileRecv;

	if (!IS_UNCACHED_SEG(&SearchFileRecv))
		SifWritebackDcache(&SearchFileRecv, sizeof(CdlFILE));

	SifCallRpc(&CdClientSearchFile, 0, 0, rpcBuffer, sizeof(struct SearchFilePkt), rpcBuffer, 4, 0, 0);

	if(rpcBuffer[0])
		memcpy(fp, &SearchFileRecv, sizeof(CdlFILE));

	return(rpcBuffer[0]);
}

static void _CdReadIntr(void *data_raw)
{
   int	*pkt = (u32 *)UNCACHED_SEG( data_raw);
   u8	*data, *out;
   int  d1, i;

   d1 = pkt[0];
   if (d1 > 0)
      {
      out = (u8 *)pkt[2];
      if (out)
         {
         i = 0;
         data = (u8 *)&pkt[4];

         do {
            i++;
            *out = *data;
            out++; data++;
            } while (i < d1);
         }
      }
   d1 = pkt[1];
   if (d1 > 0)
      {
      out = (u8 *)pkt[3];
      if (out)
         {
         i = 0;
         data = (u8 *)&pkt[8];
         do {
            i++;
            *out = *data;
            out++; data++;
            } while (i < d1);
         }
      }
}

int CdRead(u32 lbn, u32 sectors, void *buf, CdRMode *mode)
{
	struct ReadPkt * readPkt = (struct ReadPkt *)rpcBuffer;
	int numBytes = 0;

	if(!CdInited) return -1;

	readPkt->lbn = lbn;
	readPkt->sectors = sectors;
	readPkt->buffer = buf;
	memcpy(&readPkt->ReadMode, mode, sizeof(CdRMode));
	readPkt->intr_data = intrData;
	readPkt->read_pos = readCurPos;

	if(mode->datapattern == CdSecS2048)
		numBytes = sectors * 2048;
	else if(mode->datapattern == CdSecS2328)
		numBytes = sectors * 2328;
	else if(mode->datapattern == CdSecS2340)
		numBytes = sectors * 2340;

	SifWritebackDcache(buf, numBytes);

	readCurPos[0] = 0;
	SifWritebackDcache(intrData, 144);
	SifWritebackDcache(readCurPos, 4);

	if(SifCallRpc(&CdClientNcmd, CD_NCMD_READ, 0, rpcBuffer, sizeof(struct ReadPkt), rpcBuffer, 4, _CdReadIntr, intrData))
		return 0;

	return 1;
}

int CdStandby(void)
{
	if(!CdInited) return -1;
	
	SifCallRpc(&CdClientNcmd, CD_NCMD_STANDBY, 0, 0, 0, 0, 0, 0, 0);

	return 1;
}

int CdReadKey(int unk1, int unk2, int unk3, unsigned char *key)
{
	if(!CdInited) return -1;

	rpcBuffer[0] = unk1;
	rpcBuffer[1] = unk2;
	rpcBuffer[2] = unk3;

	SifCallRpc(&CdClientNcmd, CD_NCMD_READKEY, 0, rpcBuffer, 12, rpcBuffer, 20, 0, 0);

	memcpy((void *)key, (void *)&rpcBuffer[1], 16);
	return rpcBuffer[0];
}
