#include <tamtypes.h>
#include <kernel.h>
#include <stdio.h>
#include <stdlib.h>

#include "minilzo.h"

//#define DEBUG

extern unsigned char __attribute__((aligned(16))) PackedElf[];
#define BLOCK_SIZE		256*1024
static unsigned char WorkMem[BLOCK_SIZE + 8];

typedef struct {
	u32 CompressedSize;
	u32 OriginalSize;
	u32 VirtualAddr;
	u32 EntryAddr;
} SjCRUNCH_Header;

#ifdef DEBUG
int get_checksum(const char* buffer, int size)
{
	unsigned int checksum = 0;
	int i;

	for(i=0;i<size;i++)
		checksum += (unsigned char)!buffer[i];

	return checksum;
}
#endif

void Decompress(u8 *dest, u8 *src);

void *decompressLoader()
{
	SjCRUNCH_Header *Header;

	Header = (SjCRUNCH_Header *)PackedElf;

#ifdef DEBUG
	printf("CompressedSize = %d\n", Header->CompressedSize);
	printf("OriginalSize = %d\n", Header->OriginalSize);
	printf("VirtualAddr = 0x%X\n", Header->VirtualAddr);
	printf("EntryAddr = 0x%X\n", Header->EntryAddr);
#endif

	// Decompress
	Decompress((u8 *)Header->VirtualAddr, (u8 *)(PackedElf + sizeof(SjCRUNCH_Header)));

#ifdef DEBUG
	printf("Checksum = 0x%X\n", get_checksum(Header->VirtualAddr, Header->OriginalSize));
#endif

	// Return entry addr, which is jumped to
	return (void *)Header->EntryAddr;
}

void Decompress(u8 *dest, u8 *src)
{
	u8 *In, *Out;
	u32 CompressedSize;
	u32 UncompressedSize;
	u32 NewSize;
	u32 Return;

	if (lzo_init() != LZO_E_OK)
	{
#ifdef DEBUG
		printf("lzo_init FAILED!\n");
#endif
		while(1);
	}

	In = WorkMem;
	Out = dest;

	// Position pointer after identifier string
	src += 32;

	for(;;)
	{

		lzo_memcpy(In, src, BLOCK_SIZE + 8);

		UncompressedSize = *(u32 *)In;
		In += 4;
		CompressedSize = *(u32 *)In;
		In += 4;

		if(UncompressedSize == 0) break; // End of compressed data

		if(CompressedSize < UncompressedSize)
		{

			NewSize = UncompressedSize;

			Return = lzo1x_decompress( In, CompressedSize, Out, &NewSize, NULL);

			if ((Return != LZO_E_OK) || (NewSize != UncompressedSize)) // Error
			{
#ifdef DEBUG
				printf("Decompression error!!!\n");
#endif
				while(1);
			}

		} else { // In the case of uncompressed data

			lzo_memcpy(Out, In, UncompressedSize);

		}

		// Update pointers
		src += CompressedSize + 8;
		Out += UncompressedSize;

	}

}
