#include "bootmaster.h"

/////////////////////////////////////////////////////////////////////////////////////
//   Debug stuff

#define DEBUG

#ifdef DEBUG
	#define GS_BGCOLOUR	*((volatile unsigned long int*)0x120000E0)
#endif

/////////////////////////////////////////////////////////////////////////////////////
//   Global variables

int *syscallTable _RESIDENT_ = NULL;

/////////////////////////////////////////////////////////////////////////////////////
//   Local function prototypes

static unsigned int get_sr() _RESIDENT_;
static void set_sr(unsigned int v) _RESIDENT_;
static void myGsSetCrt(int inter, int mode, int interMode) _RESIDENT_;
static void myLoadExecPS2(char *filename, int argc, char **argv) _RESIDENT_;
static void iopRebootHandler() _RESIDENT_;
static void *scanIopReadKey() _RESIDENT_;
static void patchIopReturn(volatile unsigned int *address, int retval) _RESIDENT_;

void SifSetRegWatcher(int reg, int val) _RESIDENT_;

// From hax0r_asm.S
void oldSifSetReg() _RESIDENT_;
void umSifSetRegWatcher() _RESIDENT_;

void testDebugHandler() _RESIDENT_;
void yposHandler() _RESIDENT_;

/////////////////////////////////////////////////////////////////////////////////////
//   Local variables

static void (*oldsceSetGsCrt)(int inter, int mode, int interMode) _RESIDENT_;
static int sceSetGsCrtMode _RESIDENT_;
static void (*IopRebootCallback)(void) _RESIDENT_;
static int SifSetRegWatchFlag _RESIDENT_ = 0;
static void (*ElfLoaderEntry)(int argc, char **argv) _RESIDENT_;
static char *LoadExecPS2ArgBuf[16] _RESIDENT_;
int ydiffLowRes _RESIDENT_;
int ydiffHighRes _RESIDENT_;


///////////////////////////////////////////////////////////////////////////////
//   Misc hax0r code & init code

void *decompressLoader();

int initHax0r()
{
	void *entryPoint;

	// Depack custom elf loader into memory
	entryPoint = decompressLoader();

	// Find system call table
	syscallTable = findSyscallTable();

	if(syscallTable == NULL) return -1;

	// Install iop reboot watcher and handler
	installIopRebootHandler((void *)iopRebootHandler);

	// Install LoadExecPS2 replacement
	installMyLoadExecPS2(entryPoint);

	return 0;
}

_RESIDENT_ static unsigned int get_sr() {
    register unsigned int rv;
    asm volatile ("
        mfc0 %0, $12
		sync
		nop
		" : "=r" (rv) : );
    return rv;
}

_RESIDENT_ static void set_sr(unsigned int v) {
    asm volatile ("
        mtc0 %0, $12
		sync
		nop
        " : : "r" (v) );
}

_RESIDENT_ void set_kmode() {
	set_sr(get_sr() & (~0x19));
}

_RESIDENT_ void set_umode() {
	set_sr((get_sr() & (~0x19)) | 0x11);
}

///////////////////////////////////////////////////////////////////////////////
//   LoadExecPS2 patcher and replacement

void installMyLoadExecPS2(void *entryPoint)
{
	set_kmode();

	ElfLoaderEntry = entryPoint;
	syscallTable[0x06] = (int)myLoadExecPS2;

	set_umode();
} 

_RESIDENT_ static void myLoadExecPS2(char *filename, int argc, char **argv)
{
	int i;

	SifSetRegWatchFlag = 0;

	LoadExecPS2ArgBuf[0] = filename;

	if(argc > 15) argc = 15;

	for(i = 0; i < argc; i++)
		LoadExecPS2ArgBuf[i + 1] = argv[i];

	// Wont return
	ExecPS2(ElfLoaderEntry, NULL, argc + 1, LoadExecPS2ArgBuf);
}

///////////////////////////////////////////////////////////////////////////////
//   Patch out call to InitialiseUserMemory()

int disableClearUserMem()
{
	int *scanMem = (int *)0x80000000;
	int retVal = -1;
	int i;

	set_kmode();

	for(i = 0; i < 0x3FFFF; i++)
	{
		if(	((scanMem[i] & 0xffff0000) == 0x3C040000) &&		// lui a0, 0000
			((scanMem[i + 1] & 0x7C000000) == 0xC000000) &&		// jal XXXX
			(scanMem[i + 2] == 0x34842000) )					// ori a0, a0, 0x2000
			{
				scanMem[i + 1] = NOP;
				retVal = 0;
				break;
			}
	}

	set_umode();

	return retVal;
}


///////////////////////////////////////////////////////////////////////////////
//   IOP reboot detection code & callbacks

_RESIDENT_ static void *scanIopReadKey()
{
	int i;
	volatile unsigned int *iopMem = (volatile unsigned int *)IOP_MEM_BASE;
	unsigned int *patchAddress = NULL;

	for(i = 0; i < 0x0007FFFC; i++)
		if(iopMem[i] == 0x3C02BF40)
			if(iopMem[i + 1] == 0x34422020)
				if(iopMem[i + 2] == 0x3C03BF40)
					if(iopMem[i + 3] == 0x34632039)
					{
						patchAddress = (unsigned int *)&iopMem[i - 18];
						break;
					}

	return patchAddress;
}

_RESIDENT_ static void patchIopReturn(volatile unsigned int *address, int retval)
{
	address[0] = MAKE_ORI(2, 0, retval);
	address[1] = JR_RA;
	address[2] = NOP;
}


_RESIDENT_ static void iopRebootHandler()
{
	unsigned int *readkeyAddress = NULL;

	set_kmode();

	readkeyAddress = scanIopReadKey();
	if(readkeyAddress != NULL)
		patchIopReturn(readkeyAddress, 1);

	set_umode();
}

_RESIDENT_ void SifSetRegWatcher(int reg, int val)
{	

	// Make standard SifSetReg call
	oldSifSetReg(reg, val);

	// Clear flag on new IOP reboot
	if((reg == 4) && (val == 0x10000))
	{
		SifSetRegWatchFlag = 0;
		return;
	}

	// Wait until IOP reboot is complete
	if((reg == 0x80000000) && (val == 0))
	{
		SifSetRegWatchFlag = 1;
		return;
	}

	// Only do patching if called from SifInitRpc
	if((reg != 0x80000002) || (val != 1) || (SifSetRegWatchFlag == 0))
		return;

	IopRebootCallback();

#ifdef DEBUG
	GS_BGCOLOUR = 0x00ff00;
#endif
}


void installIopRebootHandler(void (*functionPtr)(void))
{
	set_kmode();
		
	syscallTable[251] = syscallTable[121];		// sceSifSetReg
	IopRebootCallback = functionPtr;
	syscallTable[121] = (int)umSifSetRegWatcher;

	set_umode();
}

///////////////////////////////////////////////////////////////////////////////
//   PAL/NTSC patcher code

_RESIDENT_ static void myGsSetCrt(int inter, int mode, int interMode) 
{	
	oldsceSetGsCrt(inter, sceSetGsCrtMode, interMode);
}

void installPNSelector(int vid_mode)
{
	// Install sceGsSetCrt patcher
	set_kmode();

	sceSetGsCrtMode = vid_mode;
	oldsceSetGsCrt = (void *)syscallTable[2];
	syscallTable[2] = (int)myGsSetCrt;

	set_umode();

	// Setup breakpoint stuff for y-pos fix
	if(vid_mode == 3)		// PAL
	{
		ydiffLowRes = 0;
		ydiffHighRes = 50;
	} 
	else 
	{
		ydiffLowRes = 20;
		ydiffHighRes = 10;
	}

	installDebugExceptionHandler((void *)yposHandler);

	// Trap writes to 0x12000080 and 0x120000A0
	setDataAddressWriteBreakpoint((void *)0x12000080, 0xFFFFFFDF); 

}

///////////////////////////////////////////////////////////////////////////////
//   System call related code

void *findSyscallTable()
{
	int *scanMem = (int *)0x80000000;
	void *tableAddress = NULL;
	int i;

	setSyscall(252, (void *)0xDEADBEEF);
	setSyscall(253, (void *)0x1337C0DE);
	setSyscall(254, (void *)findSyscallTable);

	set_kmode();

	for(i = 0; i < 0x3FFFF; i++)
		if(	(scanMem[i] == 0xDEADBEEF) &&
			(scanMem[i + 1] == 0x1337C0DE) &&
			(scanMem[i + 2] == (int)findSyscallTable) )
			{
				tableAddress = (void *)&scanMem[i - 252];
				break;
			}

	set_umode();

	return tableAddress;
}

void *getSyscall(int number)
{
	void *retVal;

	set_kmode();
	retVal = (void *)syscallTable[number];
	set_umode();

	return retVal;
}

///////////////////////////////////////////////////////////////////////////////
//   Breakpoint (y-pos fix) related code

void installDebugExceptionHandler(void (*functionPtr)(void))
{
	set_kmode();
	*((unsigned int*)0x80000100) = MAKE_J(functionPtr);
	set_umode();
}
